####################
# File name: A2.py
# This is the programming portion of Assignment 2
# for CPSC 447/547 Intro to Quantum Computing
#
# Name: 
# NetID:
####################

import requirement_A2
import math
import numpy as np

####################
# Helper functions
####################

np.random.seed(447547) 

def complexClose(d1, d2, epsilon=10**-7):
    return (abs(d1.real - d2.real) < epsilon) and (abs(d1.imag - d2.imag) < epsilon) 

def vectorClose(v1, v2, epsilon=10**-7):
    return len(v1) == len(v2) and bool(all([complexClose(v1[i], v2[i], epsilon=10**-7) for i in range(len(v1))]))

def matrixClose(m1, m2, epsilon=10**-7):
    return np.allclose(m1, m2, atol=epsilon)

def sampleBit(p):
    r = np.random.random_sample()
    if r < p:
        return bool(0)
    else:
        return bool(1)

####################
# Place for your own helper functions, if any
####################



####################
# Building Blocks
####################
        
class Qubit(object):
    """Qubit object"""
    def __init__(self, arg, label='q'):
        super(Qubit, self).__init__()
        self.arg = arg
        self.label = label

class Backend(object):
    """docstring for Backend"""
    def __init__(self, num_q, label='sys'):
        super(Backend, self).__init__()
        if (num_q == None):
            self.variable = True
            self.num_q = 0
            self.label = label
            self.in_use = 0
            self.all_qubits = []
        else:
            self.variable = False
            self.num_q = num_q
            self.label = label
            self.in_use = 0
            self.all_qubits = [Qubit(i, label) for i in range(num_q)]

    def alloc(self, n):
    	# YOUR IMPLEMENTATION HERE

        return None


class QuantumRegister(object):
    """QuantumRegister is where we keep track of qubits"""
    def __init__(self, qubits, label='qreg'):
        super(QuantumRegister, self).__init__()
        if isinstance(qubits, list):
            if ((len(qubits)==0) or (not isinstance(qubits[0], Qubit))):
                raise Exception("Invalid inputs to QuantumRegister.")
            self.size = len(qubits)
            self.label = qubits[0].label
            self.array = qubits
            # new_q is a list of indicator for whether the qubits in this register
            # is newly initialized or copied from other existing registers.
            self.new_q = [False for _ in qubits]
        elif isinstance(qubits, int):
            # New qubits in register
            self.size = qubits
            self.label = label
            self.array = [Qubit(i, label) for i in range(qubits)]
            # new_q is a list of indicator for whether the qubits in this register
            # is newly initialized or copied from other existing registers.
            self.new_q = [True for _ in range(qubits)]
        else:
            raise Exception("Invalid inputs to QuantumRegister.")

    def select(self, ids):
        # Return a new QuantumRegister by selecting a subset of qubits
        # YOUR IMPLEMENTATION HERE

        return None

    def __add__(self, other):
        # YOUR IMPLEMENTATION HERE

        return None

class ClassicalRegister(object):
    """ClassicalRegister is where we keep track of measurement outcomes"""
    def __init__(self, num_c, label='creg'):
        super(ClassicalRegister, self).__init__()
        self.size = num_c
        self.label = label
        self.state = np.array([bool(0) for _ in range(num_c)])
        
    def select(self, ids):
        # YOUR IMPLEMENTATION HERE

        return None

    def __add__(self, other):
        # YOUR IMPLEMENTATION HERE

        return None

class Gate(object):
    """Gate object to describe its name, kind, and matrix"""
    def __init__(self, name, num_q, matrix):
        super(Gate, self).__init__()
        self.name = name 
        self.num_q = num_q
        self.matrix = matrix
    def __repr__(self):
        return self.name

		
####################
# In-House Quantum Circuit
####################

class QuantumCircuit(object):
    """QuantumCircuit"""
    def __init__(self, qubits, cbits, name='module'):
        super(QuantumCircuit, self).__init__()
        self.name = name
        if isinstance(qubits, QuantumRegister):
            if ((qubits.size==0) or (not isinstance(qubits.array[0], Qubit))):
                raise Exception("Invalid inputs to QuantumCircuit.")
            self.num_q = qubits.size
            self.qubits = qubits # use the input qubits
        elif isinstance(qubits, int):
            self.num_q = qubits
            self.qubits = QuantumRegister(qubits) # initialized qubits
        else:
            raise Exception("Invalid inputs to QuantumCircuit.")
        if isinstance(cbits, ClassicalRegister):
            self.num_c = cbits.size
            self.cbits = cbits # use the input cbits
        elif isinstance(cbits, int):
            self.num_c = cbits
            self.cbits = ClassicalRegister(cbits) # initialized cbits
        else:
            raise Exception("Invalid inputs to QuantumCircuit.")
        self.circuit = [] # sequence of instructions


    def _append(self, operation, q_array, c_array):
        # Add new instruction to circuit
        instruction = [operation, q_array, c_array]
        self.circuit.append(instruction)

    def allocate(self, ids, new):
        # YOUR IMPLEMENTATION HERE
        
        return (None, None)

    def __repr__(self):
        # For displaying quantum circuit
        qasm = ['\n======<CPSC 447/547 QASM>======']
        qasm += ['Qreg: %d, Creg: %d' % (self.num_q, self.num_c)]
        for inst in self.circuit:
            #print(inst)
            (op, q_arr, c_arr) = inst
            inst_str = '%s ' % op.name 
            for q in q_arr:
                qubit = self.qubits.array[q]
                inst_str += '%s%d ' % (qubit.label, qubit.arg)
            inst_str += ', '
            for c in c_arr:
                inst_str += '%s%d ' % (self.cbits.label, c)
            qasm.append(inst_str)
        qasm.append('===============================\n')
        return "\n".join(qasm)

    # Define instruction set
    # Hadamard gate
    def h(self, qubit):
        # Define Gate by its name, kind (number of qubit), and matrix
        HGate = Gate('h', 1, None)
        # add to the end of the self.circuit list
        self._append(HGate, [qubit], [])
        return
    # Pauli X gate
    def x(self, qubit):
        XGate = Gate('x', 1, None)
        self._append(XGate, [qubit], [])
        return
    # Pauli Y gate
    def y(self, qubit):
        YGate = Gate('y', 1, None)
        self._append(YGate, [qubit], [])
        return
    # Pauli Z gate
    def z(self, qubit):
        ZGate = Gate('z', 1, None)
        self._append(ZGate, [qubit], [])
        return
    # Phase gate (sqrt(Z))
    def s(self, qubit):
        SGate = Gate('s', 1, None)
        self._append(SGate, [qubit], [])
        return
    # S dagger gate (adjoint of S gate)
    def sdg(self, qubit):
        SDGGate = Gate('sdg', 1, None)
        self._append(SDGGate, [qubit], [])
        return
    # T gate (sqrt(S))
    def t(self, qubit):
        TGate = Gate('t', 1, None)
        self._append(TGate, [qubit], [])
        return
    # T dagger gate (adjoint of T gate)
    def tdg(self, qubit):
        TDGGate = Gate('tdg', 1, None)
        self._append(TDGGate, [qubit], [])
        return
    # Controlled X gate (CNOT)
    def cx(self, ctrl_qubit, trgt_qubit):
        CXGate = Gate('cx', 2, None)
        self._append(CXGate, [ctrl_qubit, trgt_qubit], [])
        return
    # Controlled Z gate (CZ)
    def cz(self, ctrl_qubit, trgt_qubit):
        CZGate = Gate('cz', 2, None)
        self._append(CZGate, [ctrl_qubit, trgt_qubit], [])
        return
    # Toffoli gate
    def toffoli(self, ctrl_qubit_1, ctrl_qubit_2, trgt_qubit):
        TOFFGate = Gate('toffoli', 3, None)
        self._append(TOFFGate, [ctrl_qubit_1, ctrl_qubit_2, trgt_qubit], [])
        return
    # Measure qubits in array 'qubits' and store classical outcome in 'cbits'
    def measure(self, qubits, cbits):
        assert(len(qubits) == len(cbits))
        Measure = Gate('measure', len(qubits), None)
        self._append(Measure, qubits, cbits)
        return

    ####################
    # In-House Quantum Compiler
    ####################

    def compile_fully_connected(self, backend, mapping={}):
        # Flatten circuit and allocate register
        # Generate qubit mapping (qreg i to backend idx)
        # Generate the sequential circuit that runs on qubits in backend 
        # Assume the backend is fully connected here.
        # YOUR IMPLEMENTATION HERE
        
        return self.circuit # Change this


    def compile(self, backend, mapping={}):
        # Flatten circuit and allocate register
        # Generate qubit mapping (qreg i to backend idx)
        # Generate the sequential circuit that runs on qubits in backend 
        # Do not assume the backend is fully connected here
        # YOUR IMPLEMENTATION HERE
        
        return self.circuit # Change this


####################
# In-House Quantum Algorithms
####################

def moduleA(qc, qids):
    # Allocate all qubits for this module (including one new ancilla)
    (qreg, creg) = qc.allocate(qids, 1)
    # Initialize circuit for the module
    module = QuantumCircuit(qreg, creg)
    # Implementation of moduleA
    #  YOUR IMPLEMENTATION HERE



    # Add the module to the parent circuit
    qc._append(module, qids, qids)
    return

def moduleB(qc, qids):
    # Allocate all qubits for this module (including one new ancilla)
    (qreg, creg) = qc.allocate(qids, 1)
    module = QuantumCircuit(qreg, creg)
    # Implementation of moduleB
    #  YOUR IMPLEMENTATION HERE



    qc._append(module, qids, qids)
    return

def synthetic_algo(n):
    # Assume the input n is 4
    qc = QuantumCircuit(n, n)
    #  YOUR IMPLEMENTATION HERE



    return qc


####################
# Test functions (feel free to add more of your own tests here)
####################
def testCompile():
    smallQ = Backend(5, 'sys_a')
    mediumQ = Backend(50, 'sys_b')
    infinityQ = Backend(None, 'sys_c')
    qc = synthetic_algo(4)
    circ = qc.compile(mediumQ)

    qc_compiled = QuantumCircuit(50,50)
    qc_compiled.circuit = circ
    print(qc_compiled)


####################
# Main tests
####################

def testSelect1():
    q1 = QuantumRegister(4)
    qreg = q1.select([0, 2])
    assert qreg.size == 2
    assert [qubit.arg for qubit in qreg.array] == [0, 2]
    assert [qubit.label for qubit in qreg.array] == ["qreg"] * 2

def testAdd1():
    qr1 = QuantumRegister(2)
    qr2 = QuantumRegister(1)
    qreg = qr1 + qr2
    assert qreg.size == 3
    assert [qubit.arg for qubit in qreg.array] == [0, 1, 0]
    assert [qubit.label for qubit in qreg.array] == ["qreg"] * 3

def testAllocateNone():
    circ = QuantumCircuit(3, 3)
    qreg, creg = circ.allocate([0, 1, 2], 0)
    assert qreg.size == 3
    assert [qubit.arg for qubit in qreg.array] == [0, 1, 2]
    assert [qubit.label for qubit in qreg.array] == ["qreg"] * 3
    assert creg.size == 3
    assert [state for state in creg.state] == [False] * 3

def testAllocateFive():
    circ = QuantumCircuit(3, 3)
    qreg, creg = circ.allocate([0, 1, 2], 5)
    assert qreg.size == 8
    assert [qubit.arg for qubit in qreg.array] == [0, 1, 2, 0, 1, 2, 3, 4]
    assert [qubit.label for qubit in qreg.array] == ["qreg"] * 8
    assert creg.size == 8
    assert [state for state in creg.state] == [False] * 8

def testBackendAllocOnce():
    backend = Backend(5)
    backend.alloc(1)
    assert backend.num_q == 5
    assert len(backend.all_qubits) == 5
    assert backend.variable == False
    assert backend.in_use == 1
    assert [qubit.arg for qubit in backend.all_qubits] == [0, 1, 2, 3, 4]

def expandConcatenatedGateSequence(circuit):
    concatenated_gate_sequence = []
    for op, q_arr, _ in circuit:
        if isinstance(op, Gate):
            concatenated_gate_sequence.append((op.name, q_arr))
        if isinstance(op, QuantumCircuit):
            concatenated_gate_sequence.append(
                expandConcatenatedGateSequence(op.circuit)
            )
    return concatenated_gate_sequence

def testModuleBsimple():
    qc = QuantumCircuit(2, 2)
    moduleB(qc, [0])
    seq = expandConcatenatedGateSequence(qc.circuit)
    assert seq == [[('t', [0]), ('cx', [0, 1]), ('cx', [1, 0])]]

def testModuleBcomplex():
    qc = QuantumCircuit(3, 3)
    moduleB(qc, [2])
    seq = expandConcatenatedGateSequence(qc.circuit)
    assert seq == [[('t', [0]), ('cx', [0, 1]), ('cx', [1, 0])]]

def testCompileFullyConnectedSyntheticAlgoN4():
    backend = Backend(4, "moduleB")
    qc = synthetic_algo(2)
    circuit = qc.compile_fully_connected(backend)
    example_seq = [('h', [0]), ('h', [1]), ('cx', [0, 2]), ('cx', [1, 0]), ('cx', [1, 2]), ('tdg', [0]), ('t', [1]), ('cx', [1, 3]), ('cx', [3, 1])]
    if expandConcatenatedGateSequence(circuit) != example_seq:
        print("[warning] testCompileFullyConnectedSyntheticAlgoN4: mismatch with example sequence" +
            ", but your result may still be correct because the answer is not unique; try gradescope!")

def check_circuit_connectivity(circuit, adj_matrix):
    if len(circuit) == 0:
        return
    if isinstance(circuit[0], Gate):
        if len(circuit[1]) == 2:
            q0 = circuit[1][0]
            q1 = circuit[1][1]
            # print(circuit[1])
            assert adj_matrix[q0, q1] == 1, "connectivity error"
        return
    for e in circuit:
        check_circuit_connectivity(e, adj_matrix)

# copy the state to an ancilla in a hierarchy manner
# e.g. 0 -> 1, 0 -> 2, 0 -> 3...
def addHierarchy(qc, qids, depth=2):
    if depth <= 0:
        return
    assert len(qids) == 1
    # Allocate all qubits for this module
    (qreg, creg) = qc.allocate(qids, 1)
    # Initialize circuit for the module
    mqc = QuantumCircuit(qreg, creg)
    mqc.cx(0, 1)
    addHierarchy(mqc, [0], depth - 1)
    # Add the module to the parent circuit
    qc._append(mqc, qids, qids)
    return

def testCompileHierarchySimple():
    # simple adj matrix with nearest neighbor connection
    adj_matrix_simple = np.zeros((4, 4))
    def add_connection(adj_matrix, i, j):
        adj_matrix[i, j] = 1
        adj_matrix[j, i] = 1
    add_connection(adj_matrix_simple, 0, 1)
    add_connection(adj_matrix_simple, 1, 2)
    add_connection(adj_matrix_simple, 2, 3)
    # run compile
    backend = Backend(4, "Hierarchy", adj_matrix_simple)
    qc = QuantumCircuit(1, 1)
    addHierarchy(qc, [0], 3)
    circuit = qc.compile(backend)
    check_circuit_connectivity(circuit, adj_matrix_simple)
    example_seq = [('cx', [0, 1]), ('cx', [0, 1]), ('cx', [1, 0]), ('cx', [0, 1]), ('cx', [1, 2]), ('cx', [0, 1]), ('cx', [1, 0]), ('cx', [0, 1])
        , ('cx', [0, 1]), ('cx', [1, 0]), ('cx', [0, 1]), ('cx', [1, 2]), ('cx', [2, 1]), ('cx', [1, 2]), ('cx', [2, 3]), ('cx', [1, 2])
        , ('cx', [2, 1]), ('cx', [1, 2]), ('cx', [0, 1]), ('cx', [1, 0]), ('cx', [0, 1])]
    if expandConcatenatedGateSequence(circuit) != example_seq:
        print("[warning] testCompileHierarchySimple: mismatch with example sequence" +
            ", but your result may still be correct because the answer is not unique; try gradescope!")


def testAll():
    testCompile()
    # public tests
    testSelect1()  # 5.1
    testAdd1()  # 5.1
    testAllocateNone()  # 5.2
    testAllocateFive()  # 5.2
    testBackendAllocOnce()  # 5.3
    testModuleBsimple()  # 5.4
    testModuleBcomplex()  # 5.4
    testCompileFullyConnectedSyntheticAlgoN4()  # 5.5
    testCompileHierarchySimple()  # 5.6
    # ... You can add more or comment out tests

def main():
    requirement_A2.check()
    testAll()

if __name__ == '__main__':
    main()
